//#include "stdafx.h"
#include "FF3Data.h"
#include "FF3Doc.h"
#include "epff6extract.h"
#include "obbitmap.h"



WorldData::WorldData(){
	objectType = ObjWorldData;
	tiles = 0;
	ppdLastScrollPos = CPoint(0, 0);
}

WorldData::~WorldData(){
	for(int i = 0; i < triggers.GetSize(); i++){
		delete triggers[i];
	}

	if(tiles){
		delete [] tiles;
	}
}

unsigned char WorldData::ReadMapData(unsigned long dwOffset){
	DecompressData(pDoc->Fin, dwOffset, map);
	return 1;
}

unsigned char WorldData::ReadMiniMapData(unsigned long dwOffset){
	DecompressData(pDoc->Fin, dwOffset, miniMap);
	return 1;
}

unsigned char WorldData::ReadTileData(unsigned long dwOffset){
	// Extract the tiles (as an array of BYTEs) from the ROM.  They are stored
	// in a chunky 4 bits-per-pixel format.  The tile data is 8192 
	// bytes long.

	unsigned char pTiles[0x2480];
	DecompressData(pDoc->Fin, dwOffset, pTiles, 0x2480);

	//The first 0x400 bytes are TileFormations
	memcpy(&tileFormations, pTiles, 0x400);

	//The last 0x80 bytes are tilePalette shifts
	for(int i = 0; i < 256; i++)
	{
		unsigned char b1 = pTiles[0x2400 + (i / 2)];

		if(!(i & 0x01))			//if even
			tilePalettes[i] = b1 & 0xF;
		else
			tilePalettes[i] = (b1 & 0xF0) >> 4;
	}

	//The middle 2000 is actually tile data
	tiles = new unsigned char[0x2000];
	memcpy(tiles, pTiles + 0x400, 0x2000);

	return 1;
}

/*unsigned char WorldData::ReadEnemyLayout(unsigned long dwOffset, unsigned long dwLength)
{
	// Extract the tiles (as an array of BYTEs) from the ROM.  They are stored
	// in a chunky sequential 4 bits per pixel format.  The tile data is 8192 
	// bytes long.

	pDoc->Fin->Seek(dwOffset, CFile::begin);
	pDoc->Fin->Read(enemyLayout, dwLength);

	return 1;
}*/

/*unsigned char WorldData::ReadEnemyOdds(unsigned long dwOffset, unsigned long dwLength)
{
	// Extract the tiles (as an array of BYTEs) from the ROM.  They are stored
	// in a chunky sequential 4 bits per pixel format.  The tile data is 8192 
	// bytes long.

	pDoc->Fin->Seek(dwOffset, CFile::begin);
	pDoc->Fin->Read(enemyOdds, dwLength);

	return 1;
}*/

unsigned char WorldData::ReadTileProperties(unsigned long dwOffset)
{
	// Extract the information assigning various properties for each of the
	// 512 16x16 tiles.  This information tells the game engine which tiles
	// can be walked on, whether enemy encounters are activated, among other
	// things.  Each property is represented by a WORD, totalling 512 bytes
	// of data for 256 tiles.

	pDoc->Fin->Seek(dwOffset, CFile::begin);
	pDoc->Fin->Read(&tileProperties, 512);

	return 1;
}

unsigned char WorldData::ReadPalettes(unsigned long dwOffset){
	// Extract the actual color information from the ROM.  There are 128 colors
	// per map, each color being represented by a WORD in a 555 BGR format.
	// This makes a total of 256 bytes per map.

	pDoc->Fin->Seek(dwOffset, CFile::begin);
	pDoc->Fin->Read(&palette, 256);

	return 1;
}

unsigned char WorldData::ReadTriggers(unsigned long dwEntranceOffset, unsigned long dwEventOffset)
{
	unsigned long dwActualOffset = 0, dwLength = 0;
	WORD wThisOffset = 0, wNextOffset = 0;
	int numEntrances = 0, numEvents = 0;
	unsigned char bEvent[3] = { 0 };

	// Read the entrances.
	pDoc->Fin->Seek(dwEntranceOffset, CFile::begin);
	pDoc->Fin->Read(&wThisOffset, 2);
	pDoc->Fin->Read(&wNextOffset, 2);

	dwLength = wNextOffset - wThisOffset;

	numEntrances = dwLength / 6;
	triggers.SetSize(numEntrances);

	pDoc->Fin->Seek(0x1FBD00 + wThisOffset, CFile::begin);
	for(int i = 0; i < numEntrances; i++)
	{
		EntranceTrigger* pTrigger = new EntranceTrigger;
		pTrigger->type = triggerEntrance;
		pTrigger->flags = TRIGGER_MINIMAP;
		pDoc->Fin->Read(&pTrigger->srcX, 2);
		pDoc->Fin->Read(&pTrigger->destination, 4);
		triggers[i] = pTrigger;
	}

	// Now read the events.
	pDoc->Fin->Seek(dwEventOffset, CFile::begin);
	pDoc->Fin->Read(&wThisOffset, 2);
	pDoc->Fin->Read(&wNextOffset, 2);

	dwLength = wNextOffset - wThisOffset;

	numEvents = dwLength / 5;
	triggers.SetSize(numEntrances + numEvents);

	pDoc->Fin->Seek(0x040200 + wThisOffset, CFile::begin);
	for(i = numEntrances; i < numEntrances + numEvents; i++)
	{
		EventTrigger* pTrigger = new EventTrigger;
		pTrigger->type = triggerEvent;
		pTrigger->flags = TRIGGER_MINIMAP;
		pTrigger->event = 0;
		pDoc->Fin->Read(&pTrigger->srcX, 2);
		pDoc->Fin->Read(&pTrigger->event, 3);
		triggers[i] = pTrigger;
	}

	return 1;
}

Trigger* WorldData::GetTrigger(unsigned char x, unsigned char y)
{
	int i;

	for(i = 0; i < triggers.GetSize(); i++)
	{
		Trigger* pTrigger = triggers[i];
		if(pTrigger->srcX == x && pTrigger->srcY == y) return pTrigger;
	}

	return 0;
}

//////////////////////////////////////////////////////////////////////////
/*
LocationData::LocationData()
{
	objectType = ObjLocationData;
}

LocationData::~LocationData()
{
	for(int i = 0; i < entrances.GetSize(); i++)
		delete entrances[i];
}

Entrance* LocationData::FindEntranceFromSource(unsigned char x, unsigned char y)
{
	for(int i = 0; i < entrances.GetSize(); i++)
	{
		Entrance* pEntr = entrances[i];
		if(pEntr->srcX == x && pEntr->srcY == y) return pEntr;
	}

	return 0;
}

Enemy::Enemy()
{
	size.cx = size.cy = 0;
	gold = 0;
	experience = 0;
	hp = 0;
	tiles = 0;
}

Enemy::~Enemy()
{
	if(tiles) delete tiles;
}

Event::Event()
{
	size = 0;
}

Event::~Event()
{
	for(int i = 0; i < code.GetSize(); i++)
		delete code[i];
}

GlobalLocationData::GlobalLocationData()
{
	pFile = 0;
}

GlobalLocationData::~GlobalLocationData()
{
	for(int i = 0; i < 256; i++) delete maps[i];
	for(i = 0; i < 16; i++) delete tiles[i];
}

*/

CString g_dteTable[] = {
//		0,    1,    2,    3,    4,    5,    6,    7,    8,    9,    A,    B,    C,    D,    E,    F
/* 8 */ "e ", " t", ": ", "th", "t ", "he", "s ", "er", " a", "re", "in", "ou", "d ", " w", " s", "an",
/* 9 */ "o ", " h", " o", "r ", "n ", "at", "to", " i", ", ", "ve", "ng", "ha", " m", "Th", "st", "on",
/* A */ "yo", " b", "me", "y ", "en", "it", "ar", "ll", "ea", "I ", "ed", " f", " y", "hi", "is", "es",
/* B */ "or", "l ", " c", "ne", "'s", "nd", "le", "se", " I", "a ", "te", " l", "pe", "as", "ur", "u ",
/* C */ "al", " p", "g ", "om", " d", "f ", " g", "ow", "rs", "be", "ro", "us", "ri", "wa", "we", "Wh",
/* D */ "et", " r", "nt", "m ", "ma", "I'", "li", "ho", "of", "Yo", "h ", " n", "ee", "de", "so", "gh",
/* E */ "ca", "ra", "n'", "ta", "ut", "el", "! ", "fo", "ti", "We", "lo", "e!", "ld", "no", "ac", "ce",
/* F */ "k ", " u", "oo", "ke", "ay", "w ", "!!", "ag", "il", "ly", "co", ". ", "ch", "go", "ge", "e_"
};

DialogueItem::DialogueItem()
{
}

DialogueItem::~DialogueItem()
{
	if(length != 0 && text) delete [] text;
}

void DialogueItem::ConvertToASCII(CString& str, unsigned long tag, unsigned long value, ... )
{
	unsigned long propTag = tag;
	unsigned long propValue = value;
	CMap<unsigned long, unsigned long, unsigned long, unsigned long> tagColl;

	va_list marker;
	va_start(marker, value);
	while(propTag != 0)
	{
		tagColl[propTag] = propValue;
		propTag = va_arg(marker, unsigned long);
		propValue = va_arg(marker, unsigned long);
	}
	va_end(marker);

	unsigned char bTagFound = FALSE;
	unsigned long dwValue = 0;

	unsigned long dwMode = 0;
	bTagFound = tagColl.Lookup(0x83000001, dwValue);
	if(bTagFound)
		dwMode = (unsigned long)dwValue;
	else
		return;

	int nOffset = 0;
	bTagFound = tagColl.Lookup(0x83000002, dwValue);
	if(bTagFound)
		nOffset = (int)dwValue;
	else
		nOffset = 0;

	str.Empty();

	if(text == 0)
	{
		if(dwMode == DICM_List)
			str = "<Empty>";
		else
			str.Empty();

		return;
	}

	for(int i = nOffset; i < length; i++)
	{
		if(text[i] >= 0x20 && text[i] <= 0x39) str += (char) (text[i] + 0x21);
		else if(text[i] >= 0x3A && text[i] <= 0x53) str += (char) (text[i] + 0x27);
		else if(text[i] >= 0x54 && text[i] <= 0x5D) str += (char) (text[i] - 0x24);

		else if(text[i] >= 0x80 && text[i] <= 0xFF) str += g_dteTable[text[i] - 0x80];

		else if(text[i] == 0x5E) str += "!";
		else if(text[i] == 0x5F) str += "?";
		else if(text[i] == 0x60) str += "/";
		else if(text[i] == 0x61) str += ":";
		else if(text[i] == 0x62) str += "$";			// $ is the right quote in FF3.FNT
		else if(text[i] == 0x63) str += "'";
		else if(text[i] == 0x64) str += "-";
		else if(text[i] == 0x65) str += ".";
		else if(text[i] == 0x66) str += ",";
		else if(text[i] == 0x67) str += "_";			// _ is the ellipsis
		else if(text[i] == 0x68) str += ";";
		else if(text[i] == 0x69) str += "#";
		else if(text[i] == 0x6A) str += "+";
		else if(text[i] == 0x6B) str += "(";
		else if(text[i] == 0x6C) str += ")";
		else if(text[i] == 0x6D) str += "%";
		else if(text[i] == 0x6E) str += "~";
		else if(text[i] == 0x6F) str += "*";
		else if(text[i] == 0x70) str += "@";
		else if(text[i] == 0x71) str += "&";			// & is the musical note
		else if(text[i] == 0x72) str += "=";
		else if(text[i] == 0x73) str += "\"";
		else if(text[i] == 0x7F) str += " ";

		else if(text[i] == 0x01 && dwMode == DICM_List) str += " ";
		else if(text[i] == 0x01 && dwMode == DICM_Edit) str += "\r\n";
		else if(text[i] == 0x13 && dwMode == DICM_List) str += " ";
		else if(text[i] == 0x13 && dwMode == DICM_Edit) str += "<P>\r\n";
		else if(text[i] == 0x15) str += "^";			// ^ is the multiple choice indicator

		else if(text[i] == 0x19) str += "<N>";
		else if(text[i] == 0x1A) str += "<I>";
		else if(text[i] == 0x1B) str += "<S>";

		else if(text[i] == 0x02 && dwMode == DICM_List) str += "TERRA";
		else if(text[i] == 0x03 && dwMode == DICM_List) str += "LOCKE";
		else if(text[i] == 0x04 && dwMode == DICM_List) str += "CYAN";
		else if(text[i] == 0x05 && dwMode == DICM_List) str += "SHADOW";
		else if(text[i] == 0x06 && dwMode == DICM_List) str += "EDGAR";
		else if(text[i] == 0x07 && dwMode == DICM_List) str += "SABIN";
		else if(text[i] == 0x08 && dwMode == DICM_List) str += "CELES";
		else if(text[i] == 0x09 && dwMode == DICM_List) str += "STRAGO";
		else if(text[i] == 0x0A && dwMode == DICM_List) str += "RELM";
		else if(text[i] == 0x0B && dwMode == DICM_List) str += "SETZER";
		else if(text[i] == 0x0C && dwMode == DICM_List) str += "MOG";
		else if(text[i] == 0x0D && dwMode == DICM_List) str += "GAU";
		else if(text[i] == 0x0E && dwMode == DICM_List) str += "GOGO";
		else if(text[i] == 0x0F && dwMode == DICM_List) str += "UMARO";

		else if(text[i] == 0x02 && dwMode == DICM_Edit) str += "<C 1>";
		else if(text[i] == 0x03 && dwMode == DICM_Edit) str += "<C 2>";
		else if(text[i] == 0x04 && dwMode == DICM_Edit) str += "<C 3>";
		else if(text[i] == 0x05 && dwMode == DICM_Edit) str += "<C 4>";
		else if(text[i] == 0x06 && dwMode == DICM_Edit) str += "<C 5>";
		else if(text[i] == 0x07 && dwMode == DICM_Edit) str += "<C 6>";
		else if(text[i] == 0x08 && dwMode == DICM_Edit) str += "<C 7>";
		else if(text[i] == 0x09 && dwMode == DICM_Edit) str += "<C 8>";
		else if(text[i] == 0x0A && dwMode == DICM_Edit) str += "<C 9>";
		else if(text[i] == 0x0B && dwMode == DICM_Edit) str += "<C 10>";
		else if(text[i] == 0x0C && dwMode == DICM_Edit) str += "<C 11>";
		else if(text[i] == 0x0D && dwMode == DICM_Edit) str += "<C 12>";
		else if(text[i] == 0x0E && dwMode == DICM_Edit) str += "<C 13>";
		else if(text[i] == 0x0F && dwMode == DICM_Edit) str += "<C 14>";

		else if((text[i] == 0x10 || text[i] == 0x12) && dwMode == DICM_List) ;
		else if(text[i] == 0x00) ;

		else if(text[i] == 0x10 && dwMode == DICM_Edit) str += "<D>";

		else if(text[i] == 0x14 && dwMode == DICM_List)
		{
			i++;
		}
		else if(text[i] == 0x14 && dwMode == DICM_Edit)
		{
			i++;
			unsigned char param = text[i];

			CString t;  t.Format("<T %d>", param);
			str += t;
		}

		else if((text[i] == 0x11 || text[i] == 0x16) && dwMode == DICM_List)
		{
			i++;
			while(text[i] != 0x12)
				i++;			
		}
		else if(text[i] == 0x16 && dwMode == DICM_Edit)
		{
			CString t;
			i++;
			unsigned char param = text[i];

			t.Format("<WS %d", param);

			while(text[i] != 0x12)
			{
				str += t;
				i++;
				param = text[i];
				t.Format(", %d", param);
			}

//			str += t;
			str += ">";
		}
		else if(text[i] == 0x11 && dwMode == DICM_Edit)
		{
			CString t;
			i++;
			unsigned char param = text[i];

			t.Format("<WT %d", param);

			while(text[i] != 0x12)
			{
				str += t;
				i++;
				param = text[i];
				t.Format(", %d", param);
			}

//			str += t;
			str += ">";
		}

		else
		{
			CString t;
			t.Format("<%02X>", text[i]);
			str += t;
		}
	}
}



void WorldData::WriteOutMapData(char *file){
	FILE *Fout;
	int i;
	int j;
	unsigned short nTilesCount[256];
	memset(nTilesCount, 0, sizeof(nTilesCount));

	Fout = fopen(file, "w");
	
	for(i = 0; i < 256; i++){
		for(j = 0; j < 256; j++){
			nTilesCount[map[(i << 8) + j]]++;
			fprintf(Fout, "%02X", map[(i << 8) + j]);
		}
		fprintf(Fout, "\n");
	}
	for(i = 0; i < 256; i++){
		if(nTilesCount[i]){
			fprintf(Fout, "Tile%02X:%6d\n", i, nTilesCount[i]);
		}
	}
	fclose(Fout);
}

void WorldData::WriteOutTiles(char *startname){
	OBBitmap Tile;
	char filename[32];
	unsigned int x, y;

	Tile.nBitPlanes = 1;
	Tile.SetDim(256, 256);
	Tile.SetPalette(palette, 128);
	
	for(y = 0; y < Tile.nHeight; y += 16){
		for(x = 0; x < Tile.nWidth; x += 16){
			Tile.Put8x8Chunk(x, y, &tiles[tileFormations[(y | (x >> 4)) * 4] << 5], tilePalettes[tileFormations[(y | (x >> 4)) * 4]]);
			Tile.Put8x8Chunk(x + 8, y, &tiles[tileFormations[(y | (x >> 4)) * 4 + 1] << 5], tilePalettes[tileFormations[(y | (x >> 4)) * 4 + 1]]);
			Tile.Put8x8Chunk(x, y + 8, &tiles[tileFormations[(y | (x >> 4)) * 4 + 2] << 5], tilePalettes[tileFormations[(y | (x >> 4)) * 4 + 2]]);
			Tile.Put8x8Chunk(x + 8, y + 8, &tiles[tileFormations[(y | (x >> 4)) * 4 + 3] << 5], tilePalettes[tileFormations[(y | (x >> 4)) * 4 + 3]]);
		}
	}
	sprintf(filename, "maptiles\\%s.bmp", startname);
	Tile.WriteOutBitmap(filename);
}