#ifndef __FF4DATA_H__
#define __FF4DATA_H__

#include <afxcoll.h>
#include <afxtempl.h>



class CFF3Doc;
class WorldData;



enum ObjectTypes{
	ObjWorldData = 1,
	ObjLocationData
};



class FF3Object{
public:
	int objectType;
	unsigned char modified;

	FF3Object() {
		modified = 0; 
	}
};



class DialogueItem{
public:
	int length;
	unsigned char* text;
	
	DialogueItem();
	~DialogueItem();
	void ConvertToASCII(CString& str, unsigned long tag, unsigned long value, ... );
};



// DI_ConvertMode
#define DICM_List				0x00000001
#define DICM_Edit				0x00000002



enum TriggerTypes{
	triggerInvalid = 0,
	triggerEntrance,
	triggerEvent,
};

#define TRIGGER_MINIMAP		0x00000001

class Trigger{
public:
	TriggerTypes type;
	unsigned long flags;
	unsigned char srcX;
	unsigned char srcY;
};

class EntranceTrigger : public Trigger{
public:
	unsigned short destination;
	unsigned char destX;
	unsigned char destY;
};

class EventTrigger : public Trigger{
public:
	unsigned long event;
};

#define MAP_BALANCE 0
#define MAP_RUIN 1
#define MAP_TRENCH 2



class WorldData : public FF3Object{
public:
	WorldData();
	~WorldData();

public:
	void Put8x8ChunkyTile(unsigned char* pBitmap, int width, int height, int dx, int dy, unsigned char bTile);
	void WriteOutTiles(char *startname);
	void WriteOutMapData(char *file);
	CFF3Doc* pDoc;

	CString name;
	CSize size;
	unsigned char map[65536];
	unsigned char *tiles;
	unsigned char miniMap[2048];
	unsigned char tileFormations[1024];
	unsigned short tileProperties[256];
	unsigned char tilePalettes[256];
	unsigned short palette[128];

	CTypedPtrArray<CPtrArray, Trigger*> triggers;

	CDWordArray mapPtrs;
	CDWordArray miniMapPtrs;
	CDWordArray tilePtrs;
	CDWordArray triggerPtrs;

	CPoint ppdLastScrollPos;

	unsigned char ReadPalettes(unsigned long dwOffset);
	unsigned char ReadMapData(unsigned long dwOffset);
	unsigned char ReadMiniMapData(unsigned long dwOffset);
	unsigned char ReadTileData(unsigned long dwOffset);
	unsigned char ReadTileProperties(unsigned long dwOffset);
	unsigned char ReadTriggers(unsigned long dwEntranceOffset, unsigned long dwEventOffset);

	Trigger* GetTrigger(unsigned char x, unsigned char y);
};



#define WDF_CHOCODISABLE		0x0001
#define WDF_AIRSHIPDISABLE		0x0002
#define WDF_AIRSHIPSHADOW		0x000C
#define WDF_WALKDISABLE			0x0010
#define WDF_CHARACTERTRANS		0x0020
#define WDF_ENEMYENCOUNTERS		0x0040
#define WDF_UNKNOWN1			0x0080
#define WDF_BATTLEBACKGROUND	0x0700
#define WDF_UNKNOWN2			0x0800
#define WDF_UNKNOWN3			0x1000
#define WDF_ALTERNATEMUSIC		0x2000
#define WDF_UNKNOWN5			0x4000
#define WDF_UNKNOWN6			0x8000



class Character{
public:
	unsigned char initialHP;
	unsigned char initialMP;
	unsigned char battleCmds[4];
	unsigned char vigor;
	unsigned char speed;
	unsigned char stamina;
	unsigned char magicPower;
	unsigned char battlePower;
	unsigned char defense;
	unsigned char magicDefense;
	unsigned char evade;
	unsigned char magicBlock;
	unsigned char weapon;
	unsigned char shield;
	unsigned char hat;
	unsigned char armor;
	unsigned char relic1;
	unsigned char relic2;
	unsigned char level;

	unsigned char name[6];

	unsigned char gfx[0x16A0];
	unsigned char portrait[800];
	unsigned char palette;
};



class CharacterGraphic{
public:
	unsigned short offset[6];
};



class Item{
public:
	unsigned char name[13];
	unsigned char description[64];
	unsigned char props[30];		// Since parameters vary per item type
};



class Esper{
public:
	unsigned char name[8];

	unsigned char spell1Rate;
	unsigned char spell1;
	unsigned char spell2Rate;
	unsigned char spell2;
	unsigned char spell3Rate;
	unsigned char spell3;
	unsigned char spell4Rate;
	unsigned char spell4;
	unsigned char spell5Rate;
	unsigned char spell5;
	unsigned char bonus;
};



class Magic{
public:
	unsigned char name[12];

	unsigned char target;
	unsigned char elemProps;
	unsigned char effect1;
	unsigned char damageType;
	unsigned char effect2;
	unsigned char mp;
	unsigned char power;
	unsigned char unknown1;
	unsigned char successRate;
	unsigned char special;
	unsigned char cond1, cond2, cond3, cond4;
};



class BattleCommand{
public:
	unsigned char name[7];
};



class ShopData{
public:
	unsigned char title;
	unsigned char items[8];
};

/////////////////////////////////////////////////////////////////////
//
//  MonsterFormation class
//  Stores data about monster formations and positioning.
//
class MonsterFormation{
protected:
	unsigned char sizeTemplate;
	unsigned char monstersPresent;
	unsigned char monsters[6];
	unsigned char monsterPositions[6];
	unsigned char bosses;

public:
// Auxiliary data
	unsigned char aux[4];

	inline unsigned char IsPresent(int nMonst) { return (monstersPresent & (1 << nMonst)) ? 1 : 0; }
	inline unsigned char IsBoss(int nMonst) { return (bosses & (1 << nMonst)) ? 1 : 0; }

	inline void SetPresent(int nMonst, unsigned char bSet) { monstersPresent = monstersPresent & ~(1 << nMonst) | (bSet ? (1 << nMonst) : 0x00); }

	inline int GetMonster(int nMonst){
		if(IsBoss(nMonst)){
			return monsters[nMonst] + 256;
		}
		else{
			return monsters[nMonst];
		}
	}

	inline void SetMonster(int nMonst, int bMonst){
		if(bMonst >= 256){
			monsters[nMonst] = bMonst & 0xFF;
			bosses |= (1 << nMonst);
		}
		else{
			monsters[nMonst] = bMonst;
			bosses &= ~(1 << nMonst);
		}
	}

	inline int GetAppearance() { return aux[0] & 0x0F; }
	inline void SetAppearance(int nAppr) { aux[0] = (aux[0] & ~0x0F) | nAppr; }

	inline int GetSizeTemplate() { return sizeTemplate >> 4; }
	inline void SetSizeTemplate(int nTempl) { sizeTemplate = nTempl << 4; }

	inline int GetXPos(int nMonster) { return (monsterPositions[nMonster] & 0xF0) >> 4; }
	inline int GetYPos(int nMonster) { return monsterPositions[nMonster] & 0x0F; }

	inline void SetXPos(int nMonster, int nPos) { monsterPositions[nMonster] = (monsterPositions[nMonster] & 0x0F) | (nPos << 4); }
	inline void SetYPos(int nMonster, int nPos) { monsterPositions[nMonster] = (monsterPositions[nMonster] & 0xF0) | nPos; }
};



class MonsterTemplate{
public:
	unsigned short unknown;
	unsigned char width;
	unsigned char height;

	inline int GetWidth() { return width; }
	inline int GetHeight() { return height; }

	inline void SetWidth(int n) { width = n; }
	inline void SetHeight(int n) { height = n; }

	inline unsigned short GetUnknown() { return unknown; }
	inline void SetUnknown(unsigned short w) { unknown = w; }
};



class MonsterVisuals{
public:
	unsigned short gfx;
	unsigned char extra;
	unsigned char palette;
	unsigned char composition;

	inline unsigned char Is3Bitplane(){ 
		return ((gfx >> 15) & 0x01); 
	}
	
	inline void SetPalette(int nPalette){
		palette = nPalette;
	}
};



class MonsterComposition{
public:
	unsigned char length;
	unsigned char mask[32];
};



class Monster{
public:
	unsigned char name[10];
	unsigned char attackName[10];

	unsigned char itemsStolen[2];
	unsigned char itemsDropped[2];

	unsigned char strength;
	unsigned char speed;
	unsigned char hitRate;
	unsigned char evadeRate;
	unsigned char magBlockRate;
	unsigned char defense;
	unsigned char magicDefense;
	unsigned char magicPower;
	unsigned short hp;
	unsigned short mp;
	unsigned short exp;
	unsigned short gold;
	unsigned char level;
	unsigned char morph;
	unsigned char flags1;
	unsigned char flags2;
	unsigned char ImmuneStatus1;
	unsigned char ImmuneStatus2;
	unsigned char ImmuneStatus3;
	unsigned char elemAbsorb;
	unsigned char elemNoEffect;
	unsigned char elemWeak;
	unsigned char attackType;
	unsigned char conditions[4];
	unsigned char specialAttack;

	int GetMorphPackage() { return morph & 0x1F; }
	int GetMorphSuccess() { return (morph & 0xE0) >> 5; }
};

#endif